local GearEvaluator = _G.GearEvaluator
local GearEvaluator_Searcher = GearEvaluator:NewModule("Searcher", "AceTimer-3.0", "AceEvent-3.0", "AceBucket-3.0")

GearEvaluator_Searcher.slotNames = {"HeadSlot",
  "NeckSlot",
  "ShoulderSlot",
  "BackSlot",
  "ChestSlot",
  "ShirtSlot",
  "TabardSlot",
  "WristSlot",
  "HandsSlot",
  "WaistSlot",
  "LegsSlot",
  "FeetSlot",
  "Finger0Slot",
  "Finger1Slot",
  "Trinket0Slot",
  "Trinket1Slot",
  "MainHandSlot",
  "SecondaryHandSlot",
  "RangedSlot",
  "AmmoSlot",
}

GearEvaluator_Searcher.invSlots = {
  ["INVTYPE_AMMO"] = "AmmoSlot",
  ["INVTYPE_HEAD"] = "HeadSlot",
  ["INVTYPE_NECK"] = "NeckSlot",
  ["INVTYPE_SHOULDER"] = "ShoulderSlot",
  ["INVTYPE_BODY"] = "ShirtSlot",
  ["INVTYPE_CHEST"] = "ChestSlot",
  ["INVTYPE_ROBE"] = "ChestSlot",
  ["INVTYPE_WAIST"] = "WaistSlot",
  ["INVTYPE_LEGS"] = "LegsSlot",
  ["INVTYPE_FEET"] = "FeetSlot",
  ["INVTYPE_WRIST"] = "WristSlot",
  ["INVTYPE_HAND"] = "HandsSlot",
  ["INVTYPE_FINGER"] = "Finger0Slot",
  ["INVTYPE_TRINKET"] = "Trinket0Slot",
  ["INVTYPE_CLOAK"] = "BackSlot",
  ["INVTYPE_WEAPON"] = "WeaponSlot",
  ["INVTYPE_SHIELD"] = "ShieldSlot",
  ["INVTYPE_2HWEAPON"] = "WeaponSlot",
  ["INVTYPE_WEAPONMAINHAND"] = "WeaponSlot",
  ["INVTYPE_WEAPONOFFHAND"] = "ShieldSlot",
  ["INVTYPE_HOLDABLE"] = "ShieldSlot",
  ["INVTYPE_RANGED"] = "RangedSlot",
  ["INVTYPE_THROWN"] = "RangedSlot",
  ["INVTYPE_RANGEDRIGHT"] = "RangedSlot",
  ["INVTYPE_RELIC"] = "RangedSlot",
  ["INVTYPE_TABARD"] = "TabardSlot",
}

GearEvaluator_Searcher.gearGroups = {
  ["HeadSlot"] = "armor",
  ["NeckSlot"] = "armor",
  ["ShoulderSlot"] = "armor",
  ["BackSlot"] = "armor",
  ["ChestSlot"] = "armor",
  ["ShirtSlot"] = "armor",
  ["TabardSlot"] = "armor",
  ["WristSlot"] = "armor",
  ["HandsSlot"] = "armor",
  ["WaistSlot"] = "armor",
  ["LegsSlot"] = "armor",
  ["FeetSlot"] = "armor",
  ["Finger0Slot"] = "rings",
  ["Finger1Slot"] = "rings",
  ["Trinket0Slot"] = "trinkets",
  ["Trinket1Slot"] = "trinkets",
  ["WeaponSlot"] = "weapons",
  ["ShieldSlot"] = "offhands",
  -- These two can be considered as armor
  ["RangedSlot"] = "armor",
  ["AmmoSlot"] = "armor",
}

GearEvaluator_Searcher.unbound = {
  ["Binds when equipped"] = true,
  ["Binds when picked up"] = true
}

function GearEvaluator_Searcher:OnInitialize()
  local defaults = {
    profile = {
      interval = 0.5,
      generations = 10,
      individuals = 20,
      survivors = 1
    }
  }

  self.db = GearEvaluator.db:RegisterNamespace("Searcher", defaults)
end

function GearEvaluator_Searcher:OnEnable()
  self:RegisterEvent("PLAYER_REGEN_ENABLED", "HandleRegenEvent")
  self:RegisterEvent("PLAYER_REGEN_DISABLED", "HandleRegenEvent")
  self:RegisterEvent("PLAYER_FLAGS_CHANGED", "HandleFlagsChangedEvent")
  self.invChangeBucket = self:RegisterBucketEvent("UNIT_INVENTORY_CHANGED", 0.1, "HandleInventoryChangedEvent")
  self.bagUpdateBucket = self:RegisterBucketEvent("BAG_UPDATE", 0.1, "HandleBagUpdateEvent")

  if (not UnitAffectingCombat("player")) then
    self:StartSearching()
  end
end

function GearEvaluator_Searcher:OnDisable()
  self:StopSearching()
  self:UnregisterEvent("PLAYER_REGEN_ENABLED")
  self:UnregisterEvent("PLAYER_REGEN_DISABLED")
  self:UnregisterEvent("PLAYER_FLAGS_CHANGED")
  self:UnregisterBucket(self.invChangeBucket)
  self:UnregisterBucket(self.bagUpdateBucket)
end

function GearEvaluator_Searcher:HandleRegenEvent(event)
  if (event == "PLAYER_REGEN_ENABLED") then
    self:StartSearching()
  else
    self:StopSearching()
  end
end

function GearEvaluator_Searcher:HandleInventoryChangedEvent(units)
  if (units.player) then self:BuildItemTable() end
end

function GearEvaluator_Searcher:HandleBagUpdateEvent()
  self:BuildItemTable()
end

function GearEvaluator_Searcher:HandleFlagsChangedEvent(event, unit)
  if (unit == "player") then
    self:StopSearching()
    self:StartSearching()
  end
end

function GearEvaluator_Searcher:StartSearching()
  if (not self.searchTimer and not UnitAffectingCombat("player") and UnitIsAFK("player")) then
    GearEvaluator:Debug(4, "Starting to search for better gear...")
    self.searchTimer = self:ScheduleRepeatingTimer("DoSearchBurst", self.db.profile.interval)
  end
end

function GearEvaluator_Searcher:StopSearching()
  if (self.searchTimer) then
    self:CancelTimer(self.searchTimer, true)
    self.searchTimer = nil
    GearEvaluator:Debug(4, "Stopped searching for better gear")
  end
end

function GearEvaluator_Searcher:AddItemToItemTable(tbl, item)
  local slot = self:GetEquipLocation(item)

  if (slot) then
    GearEvaluator:Debug(5, item.." : "..slot)

    local group = self.gearGroups[slot]

    if (group) then
      if (not tbl[group]) then
        tbl[group] = {}
      end

      if (not tbl[group][slot]) then
        tbl[group][slot] = {}
      end

      table.insert(tbl[group][slot], item)
    end
  end
end

function GearEvaluator_Searcher:BuildItemTable()
  self:StopSearching()

  -- Start from empty gear table
  self.gearTable = {}

  -- Create scanning tooltip if not available
  -- NOTE: Handling the tooltip here make things easier
  if (not self.tooltip) then
    self.tooltip = CreateFrame("GameTooltip")
    self.tooltip:SetOwner(WorldFrame, "ANCHOR_NONE")

    local lefts, rights = {}, {}

    for i = 1, 30 do
      local left, right = self.tooltip:CreateFontString(), self.tooltip:CreateFontString()
      left:SetFontObject(GameFontNormal)
      right:SetFontObject(GameFontNormal)
      self.tooltip:AddFontStrings(left, right)
      lefts[i], rights[i] = left, right
    end

    self.tooltip.lefts, self.tooltip.rights = lefts, rights
  end

  for bag = 0, 4 do
    for bagslot = 1, GetContainerNumSlots(bag) do
      local item = GetContainerItemLink(bag, bagslot)
      self.tooltip:ClearLines()
      self.tooltip:SetBagItem(bag, bagslot)
      self:AddItemToItemTable(self.gearTable, item)
    end
  end

  for slot, item in pairs(GearEvaluator:GetCurrentEquipment()) do
    self.tooltip:ClearLines()
    self.tooltip:SetInventoryItem("player", GetInventorySlotInfo(slot))
    self:AddItemToItemTable(self.gearTable, item)
  end

  self:StartSearching()
end

function GearEvaluator_Searcher:GetEquipLocation(item)
  if (not IsEquippableItem(item)) then return nil end

  for i = 1, self.tooltip:NumLines() do
    if (self.unbound[self.tooltip.lefts[i]:GetText()]) then
      GearEvaluator:Debug(2, item.." is unbound, ignoring")
      return nil
    end
  end

  local _, _, _, _, _, _, _, _, loc, _ = GetItemInfo(item)

  local slot = self.invSlots[loc]

  return slot, loc
end

function GearEvaluator_Searcher:CreateMutations(population)
  local individuals = #population
  local mutation = {}
  local slot

  -- TODO: Expand this to support other gear groups
  repeat
    slot = self.slotNames[math.random(#self.slotNames)]
  until (self.gearGroups[slot] == "armor" and self.gearTable.armor[slot] and #self.gearTable.armor[slot] > 0)

  for k, v in pairs(self.slotNames) do
    if (v == slot) then
      mutation[v] = self.gearTable.armor[v][math.random(#self.gearTable.armor[v])]
    else
      mutation[v] = population[math.random(individuals)][v]
    end
  end

  table.insert(population, mutation)
end

function GearEvaluator_Searcher:SelectSurvivors(population, evalname)
  local success
  local scores = {}

  for _, v in pairs(population) do
    success, scores[v] = GearEvaluator:GetScore(evalname, v)
  end

  table.sort(population, function(a, b)
    return scores[a] < scores[b]
  end)

  while (#population > self.db.profile.survivors) do
    table.remove(population, 1)
  end
end

function GearEvaluator_Searcher:DoSearchBurst()
  if (not self.gearTable) then return end

  if (not self.population) then
    self.population = {}
  end

  for g = 1, self.db.profile.generations do
    for k in pairs(GearEvaluator.evaluators) do
      if (not self.population[k]) then
        self.population[k] = {}
        table.insert(self.population[k], GearEvaluator:GetCurrentEquipment())
      end

      self:CreateMutations(self.population[k])
      self:SelectSurvivors(self.population[k], k)
    end
  end
end
